/***************************************************************************
                          registry.h  -  Methods for accessing the Linux Registry
                             -------------------
    begin                : Mon Dec 29 2003
    copyright            : (C) 2003 by Avi Alkalay
    email                : avi@unix.sh
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



/***************************************************************************
 *                                                                         *
 *   This is the implementation of a filesystem backend for the            *
 *   Linux Registry. Each Key is a file in the filesystem.                 *
 *   It is as secure as filesystem security. It is as reliable             *
 *   as filesystem. It uses only standards C calls, which makes it         *
 *   be usable by very low level or early boot stage software, like        *
 *   /sbin/init.  This is good business.                                   *
 *                                                                         *
 ***************************************************************************/



#include "registry.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <pwd.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <iconv.h>
#include <locale.h>
#include <langinfo.h>



#define UTF8_TO   1
#define UTF8_FROM 0

extern int errno;


int registryOpen() {
	return 0;
}


int registryClose() {
	return 0;
}


size_t keyGetSerializedSize(Key *key) {
	size_t size,tmp;


	size=sizeof(RG_KEY_FORMAT)+sizeof(u_int8_t)+1;
	if (key->comment) size+=strblen(key->comment);
	size++;
	tmp=keyGetDataSize(key);
	size+=tmp;
	return size;
}




size_t unencode(char *encoded,void *returned) {
	char byteInHexa[5]="0x";
	char *readCursor=encoded;
	char *writeCursor=returned;

	if (!encoded) {
		if (returned) *(char *)returned=0;
		return 0;
	}

	byteInHexa[4]=0;
	while (*readCursor) {
		if (isspace((int)*readCursor)) readCursor++;
		if (isxdigit((int)*readCursor)) {
			long int converted;
			byteInHexa[2]=readCursor[0];
			byteInHexa[3]=readCursor[1];
			converted=strtol(byteInHexa,0,16); /* convert from hexa to a byte */
			*writeCursor=(unsigned char)converted;

			readCursor+=2;
			writeCursor++;
		}
	}
	return (long int)writeCursor-(long int)returned;
}



int registryNeedsUTF8Conversion() {
	setlocale(LC_ALL,"");
	return strcmp(nl_langinfo(CODESET),"UTF-8");
}


/** Converts string to (direction=UTF8_TO) and
    from (direction=UTF8_FROM) UTF-8 */
int UTF8Engine(int direction, char **string, size_t *inputByteSize) {
	char *currentCharset=0;
	char *converted=0;
	char *readCursor, *writeCursor;
	size_t bufferSize;
	iconv_t converter;

	if (registryNeedsUTF8Conversion()) currentCharset=nl_langinfo(CODESET);
	else return 0;

	if (direction) converter=iconv_open("UTF-8",currentCharset);
	else converter=iconv_open(currentCharset,"UTF-8");

	if (converter == (iconv_t)(-1)) return -1;

	bufferSize=*inputByteSize*2; /* work with worst case */
	converted=malloc(bufferSize);
	if (!converted) return -1;

	readCursor=*string;
	writeCursor=converted;
	if (iconv(converter,
			&readCursor,inputByteSize,
			&writeCursor,&bufferSize) == (size_t)(-1)) {
		free(converted);
		iconv_close(converter);
		return -1;
	}

	/* calculate the UTF-8 string byte size, that will be returned */
	*inputByteSize=writeCursor+1-converted;
	/* store the current unencoded string for future free */
	readCursor=*string;
	/* allocate an optimal size area to store the converted string */
	*string=malloc(*inputByteSize);
	/* copy all that matters for returning */
	memcpy(*string,converted,*inputByteSize);
	/* release memory used by passed string */
	free(readCursor);
	/* release buffer memory */
	free(converted);
	/* release the conversor engine */
	iconv_close(converter);

	return 0;
}






int keyFileUnserialize(Key *key,FILE *input) {
	char generalBuffer[100];
	size_t currentBufferSize;

	char version[10];
	char type[5];
	char *data=0;
	size_t dataSize=0;
	char *comment=0;
	size_t commentSize=0;

	int readComment=1;
	int eof=0;

	/* The serialized format is
	   -------------------------
	   RG001\n
	   type\n
	   comment (with newlines)\n
	   <DATA>\n
	   The data encoded as text
	   -------------------------
	*/

	/* TODO: Handle different format versions */

	/* Read and currently discard the format version */
	if (!fgets(version, sizeof(version), input)) return -1;
	if (!fgets(type,    sizeof(type),    input)) return -1;

	while (readComment) {
		char *status;

		if (fgets(generalBuffer,sizeof(generalBuffer),input)) {
			if (memcmp(generalBuffer,"<DATA>\n\0",8)) {
				/* This is not the begining of the data part so it is part of comment */
				currentBufferSize=strlen(generalBuffer);
				if (!comment) {
					comment=(char *)malloc(commentSize=(currentBufferSize+1));
					strcpy(comment,generalBuffer);
				} else {
					char *buffer=0;

					buffer=malloc(commentSize+currentBufferSize);
					strcpy(buffer,comment);
					strcat(buffer,generalBuffer);
					comment=realloc(comment,commentSize+=currentBufferSize);
					strcpy(comment,buffer);
					free(buffer);
				}
			} else readComment=0;
		} else {
			readComment=0;
			eof=1;
		}
	}

	if (comment && UTF8Engine(UTF8_FROM,&comment,&commentSize)) {
		free(comment);
		return -1;
	}

	/* Now read the data section */
	if (!eof) {
		while (fgets(generalBuffer,sizeof(generalBuffer),input)) {
			currentBufferSize=strlen(generalBuffer);
			if (!data) {
				data=(char *)malloc(dataSize=(currentBufferSize+1));
				strcpy(data,generalBuffer);
			} else {
				char *buffer=0;

				buffer=malloc(dataSize+currentBufferSize);
				strcpy(buffer,data);
				strcat(buffer,generalBuffer);
				data=realloc(data,dataSize+=currentBufferSize);
				strcpy(data,buffer);
				free(buffer);
			}
		}
	}

	/* Put in the Key object */
	keySetComment(key,comment);
	keySetType(key,atoi(type));
	if (!dataSize) {
		keySetRaw(key,0,0);
		return 0;
	}

	if (keyGetType(key) <= RG_KEY_TYPE_BINARY) {
		/* Binary data. Unencode. */
		char *unencoded=0;
		size_t unencodedSize;

		/* raw data is maximum half the size of text-encoded data */
		unencodedSize=dataSize/2;

		unencoded=malloc(unencodedSize);
			keySetRaw(key,unencoded,unencode(data,unencoded));
		free(unencoded);
	} else {
		if (UTF8Engine(UTF8_FROM,&data,&dataSize)) {
			free(data);
			return -1;
		}
		keySetRaw(key,data,dataSize);
	}

	free(data);

	return 0;
}



size_t encode(void *unencoded, size_t size, char *returned) {
	void *readCursor=unencoded;
	void *writeCursor=returned;
	int blockStep=4; /* 4 bytes per block */
	int lineStep=8*blockStep; /* 8 blocks per line */
	int currentInBlock=0;
	int currentInLine=0;

	char byte[2];

	while ((readCursor-unencoded)<size) {
		sprintf(writeCursor,"%01x",*(u_int8_t *)readCursor);
		readCursor++;
		writeCursor+=2;
		currentInBlock++;
		currentInLine++;
		if (currentInLine==lineStep) {
			*(char *)writeCursor='\n'; writeCursor++;
			currentInLine=0;
			currentInBlock=0;
		}
		if (currentInBlock==blockStep) {
			*(char *)writeCursor=' '; writeCursor++;
			currentInBlock=0;
		}
	}
	*(char *)writeCursor='\n';
	*(char *)++writeCursor=0;
	return (writeCursor)-(void *)returned;
}




int keyFileSerialize(Key *key, FILE *output) {
	/* The serialized format is
	   -------------------------
	   RG001\n
	   type\n
	   comment (with newlines)\n
	   <DATA>\n
	   The data encoded as text
	   -------------------------
	*/

	size_t dataSize;

	fprintf(output,"%s\n",RG_KEY_FORMAT);
	fprintf(output,"%d\n",key->type);
	/* TODO: UTF-8 */
	if (key->comment) {
		if (registryNeedsUTF8Conversion()) {
			size_t convertedCommentSize=key->commentSize;
			char *convertedComment=malloc(convertedCommentSize);

			memcpy(convertedComment,key->comment,key->commentSize);
			if (UTF8Engine(UTF8_TO,&convertedComment,&convertedCommentSize)) {
				free(convertedComment);
				return -1;
			}
			fprintf(output,"%s\n",convertedComment);
			free(convertedComment);
		} else fprintf(output,"%s\n",key->comment);
	}

	fputs("<DATA>\n",output);
	fflush(output);

	dataSize=keyGetDataSize(key);
	if (dataSize) {
		/* There is some data to write */
		if (keyGetType(key) > RG_KEY_TYPE_BINARY) {
			if (registryNeedsUTF8Conversion()) {
				size_t convertedDataSize=key->dataSize;
				char *convertedData=malloc(convertedDataSize);

				memcpy(convertedData,key->data,key->dataSize);
				if (UTF8Engine(UTF8_TO,&convertedData,&convertedDataSize)) {
					free(convertedData);
					return -1;
				}
				fprintf(output,"%s\n",convertedData);
				free(convertedData);
			} else fputs(key->data,output);
		} else {
			char *encoded=malloc(3*dataSize);
			size_t encodedSize;

			encodedSize=encode(key->data,dataSize,encoded);
			fwrite(encoded,encodedSize,1,output);
			free(encoded);
		}
	}
	return 0;
}



size_t keyCalcRelativeFileName(Key *key,char *relativeFileName,size_t maxSize) {
	char *cursor;
	size_t pos=0;

	if (!key || !keyIsInitialized(key)) {
		errno=RG_KEY_RET_UNINITIALIZED;
		return -1;
	}
	if (!key->key) {
		errno=RG_KEY_RET_NOKEY;
		return -1;
	}

	cursor=key->key;
	while (*cursor) {
		if (pos+1 > maxSize) {
			errno=E2BIG;
			return -1;
		}
		switch (*cursor) {
			case '\\':
				cursor++;
				relativeFileName[pos]=*cursor;
				break;
			case '.':
				relativeFileName[pos]='/';
				break;
/*			case '-':
				if (*(cursor-1)=='.') {
					if (*(cursor+1)=='.')
				} */
			default:
				relativeFileName[pos]=*cursor;
		}
		cursor++;
		pos++;
	}
	relativeFileName[pos]=0;
	pos++;

	if (registryNeedsUTF8Conversion()) {
		size_t convertedSize=pos;
		char *converted=malloc(maxSize);

		memcpy(converted,relativeFileName,convertedSize);

		if (UTF8Engine(UTF8_TO,&converted,&convertedSize)) return -1;
		if (convertedSize>maxSize) {
			errno=E2BIG;
			return -1;
		}
		pos=convertedSize;
		memcpy(relativeFileName,converted,convertedSize);
		free(converted);
	}
	return pos;
}



int keyFromStat(Key *key,struct stat *stat) {
	if (!key) {
		errno=RG_KEY_RET_NULLKEY;
		return -1;
	}

	keySetAccess(key,stat->st_mode);
	keySetUID(key,stat->st_uid);
	keySetGID(key,stat->st_gid);
	if (S_ISDIR(stat->st_mode)) keySetType(key,RG_KEY_TYPE_DIR);
	key->atime=stat->st_atime;
	key->mtime=stat->st_mtime;
	key->ctime=stat->st_ctime;
	key->recordSize=stat->st_size;
	return 0;
}




size_t registryGetFilename(Key *forKey,char *returned,size_t maxSize) {
	size_t length=0;

	switch (keyGetNameSpace(forKey)) {
		case RG_NS_SYSTEM: {
			/* Prepare to use the 'system.*' database */
			strncpy(returned,RG_DB_SYSTEM,maxSize);
			length=strlen(returned);
			break;
		}
		case RG_NS_USER: {
			/* Prepare to use the 'user:????.*' database */
			char userName[100];
			struct passwd *user=0;

			length=keyGetOwner(forKey,userName,sizeof(userName));
			if (!length) strncpy(userName,getenv("USER"),sizeof(userName));

			user=getpwnam(userName);
			length=snprintf(returned,maxSize,"%s/%s",user->pw_dir,RG_DB_USER);
			break;
		} 
		default: {
			errno=RG_KEY_RET_INVALIDKEY;
			return 0;
		}
	}

	returned[length]='/'; length++;
	length+=keyCalcRelativeFileName(forKey,returned+length,maxSize-length);

	return length;
}



int registrySetValue(char *keyname, char *value) {
	Key key;
	int rc;

	keySetName(&key,keyname);
	rc=registryGetKey(&key);
	keySetString(&key,value);
	rc=registrySetKey(&key);
	keyClose(&key);
	return rc;
}



int registryGetValue(char *keyname, char *returned,size_t maxSize) {
	Key key;
	int rc=0;

	keyInit(&key);
	keySetName(&key,keyname);
	rc=registryGetKey(&key);
	if (rc == 0) keyGetString(&key,returned,maxSize);
	else rc=errno; /* store errno before a possible change */
	keyClose(&key);
	errno=rc;
	return rc;
}

/*

int registrySetInteger(char *keyname, RG_DWORD value) {
	Key key;
	int rc;

	keySetName(&key,keyname);
	rc=registryGetKey(&key);
	keySetInteger(&key,value);
	rc=registrySetKey(&key);
	keyClose(&key);
	return rc;
}



int registryGetInteger(char *keyname, RG_DWORD *returned) {
	Key key;
	int rc=0;

	keySetName(&key,keyname);
	rc=registryGetKey(&key);
	if (rc == 0) keyGetInteger(&key,returned);
	else rc=errno;
	keyClose(&key);
	errno=rc;
	return rc;
}
*/


int registryGetChildKeys(char *parentName, KeySet *returned, unsigned long options) {
	char *realParentName=0;
	size_t parentNameSize;
	DIR *parentDir;
	Key parentKey;
	char buffer[800];
	struct dirent *entry;

	/*
		- Convert parent key name into a real filename
		- Check if it is a directory. Open it
		- Browse, read and include in the KeySet
	*/
	keyInit(&parentKey);
	keySetName(&parentKey,parentName);
	registryGetFilename(&parentKey,buffer,sizeof(buffer));
	parentDir=opendir(buffer);
	if (!parentDir) return -1; /* Key is not a directory or doesn't exist. Propagate errno */

	realParentName=realloc(realParentName,parentNameSize=keyGetFullNameSize(&parentKey));
	keyGetFullName(&parentKey,realParentName,parentNameSize);

	while (entry=readdir(parentDir)) {
		Key *keyEntry;
		char *transformedName=0;
		char *cursor=entry->d_name;
		char *lastCursor=cursor;
		size_t keyNameSize=0;
		
		/* Ignore '.' and '..' directory entries */
		if (!strcmp(entry->d_name,".") || !strcmp(entry->d_name,"..")) continue;
		
		/* Dots ('.') in key name must be escaped */
		while (cursor=index(cursor,'.')) {
			if (!transformedName) transformedName=realloc(transformedName,200);
			strncat(transformedName,lastCursor,cursor-lastCursor);
			strncat(transformedName,"\\",1);
			lastCursor=cursor;
			cursor++;
		} 
		/* copy the last chunk */
		if (transformedName) {
			strcat(transformedName,lastCursor);
			keyNameSize=strblen(transformedName);
		}
		
		
		/* Next 2 ifs are required to transform from UTF-8 */
		if (!transformedName) {
			transformedName=realloc(transformedName,keyNameSize=strblen(entry->d_name));
			strcpy(transformedName,entry->d_name);
		}
		if (UTF8Engine(UTF8_FROM,&transformedName,&keyNameSize)) {
			free(transformedName);
			return -1;
		}
		
		/* Copy the entire transformed key name to our final buffer */
		sprintf(buffer,"%s.%s",realParentName,transformedName);
		free(transformedName); /* don't need it anymore */
		
		keyEntry=(Key *)malloc(sizeof(Key)); keyInit(keyEntry);
		keySetName(keyEntry,buffer);
		
		if (options & RG_O_STATONLY) registryStatKey(keyEntry);
		else {
			int rc=registryGetKey(keyEntry);
			/* If this is a permission problem, at least stat the key */
			if (rc && errno==RG_KEY_RET_NOCRED) registryStatKey(keyEntry);
		}
		
		if (S_ISDIR(keyEntry->access)) {
			if (options & RG_O_RECURSIVE) {
				KeySet children;
				char *fullName;
				size_t fullNameSize;

				fullName=malloc(fullNameSize=keyGetFullNameSize(keyEntry));
				keyGetFullName(keyEntry,fullName,fullNameSize);

				ksInit(&children);
				registryGetChildKeys(fullName,&children,options);
				
				/* Insert the current directory key in the returned list before its children */
				if (options & RG_O_DIR) ksAppend(returned,keyEntry);
				else {
					keyClose(keyEntry);
					free(keyEntry);
				}
				
				/* Insert the children */
				ksAppendKeys(returned,&children);
				free(fullName);
			} else if (options & RG_O_DIR) ksAppend(returned,keyEntry);
				else {
					keyClose(keyEntry);
					free(keyEntry);
				} 
		} else if (!(options & RG_O_NOVALUE)) ksAppend(returned,keyEntry);
	} /* while(readdir) */
	keyClose(&parentKey);
	free(realParentName);

	return 0;
}



// int registryStatChildKeys(char *parentName, KeySet *returned, unsigned long options) {
// 	DIR *parentDir;
// 	Key parentKey;
// 	char buffer[800];
// 	struct dirent *entry;
// 
// 	/*
// 		- Convert parent key name into a real filename
// 		- Check if it is a directory. Open it
// 		- Browse, read and include in the KeySet
// 	*/
// 	keyInit(&parentKey);
// 	keySetName(&parentKey,parentName);
// 	registryGetFilename(&parentKey,buffer,800);
// 	parentDir=opendir(buffer);
// 	if (!parentDir) return -1; /* propagate errno */
// 
// 	while (entry=readdir(parentDir)) {
// 		Key *keyEntry;
// 		char transformedName[800]="";
// 		char *cursor=entry->d_name;
// 		char *lastCursor=cursor;
// 		
// 		/* Ignore '.' and '..' directory entries */
// 		if (!strcmp(entry->d_name,".") || !strcmp(entry->d_name,"..")) continue;
// 		
// 		while (cursor=index(cursor,'.')) {
// 			/* Dots ('.') in key name must be escaped */
// 			strncat(transformedName,lastCursor,cursor-lastCursor);
// 			strncat(transformedName,"\\",1);
// 			lastCursor=cursor;
// 			cursor++;
// 		} 
// 		if (transformedName[0]) {
// 			/* copy the last chunk */
// 			strcat(transformedName,lastCursor);
// 			
// 			/* Copy the entire transformed key name */
// 			sprintf(buffer,"%s.%s",realParentName,transformedName);
// 		} else sprintf(buffer,"%s.%s",realParentName,entry->d_name);
// 		
// 		keyEntry=(Key *)malloc(sizeof(Key)); keyInit(keyEntry);
// 		keySetName(keyEntry,buffer);
// 		
// 		if (options & RG_O_STATONLY) registryStatKey(keyEntry);
// 		else registryGetKey(keyEntry);
// 		
// 		ksAppend(returned,keyEntry);
// 		if ((options & RG_O_RECURSIVE) && S_ISDIR(keyEntry->access)) {
// 			KeySet children;
// 			char *fullName;
// 			size_t fullNameSize;
// 
// 			fullName=malloc(fullNameSize=keyGetFullNameSize(keyEntry));
// 			keyGetFullName(keyEntry,fullName,fullNameSize);
// 
// 			ksInit(&children);
// 			registryGetChildKeys(fullName,&children,recursive);
// 			ksAppendKeys(returned,&children);
// 			free(fullName);
// 		}
// 	}
// 	keyClose(&parentKey);
// 
// 	return 0;
// }



int registryGetRootKeys(KeySet *returned) {
	Key *system,*user;
	
	system=malloc(sizeof(Key));
	keyInit(system);
	keySetName(system,"system");
	if (registryGetKey(system)) {
		free(system);
		system=0;
	}
	
	user=malloc(sizeof(Key));
	keyInit(user);
	keySetName(user,"user");
	if (registryGetKey(user)) {
		free(user);
		user=0;
	}
	
	if (system) ksInsert(returned,system);
	if (user) ksAppend(returned,user);
	
	return 0;
}



int registryStatKey(Key *key) {
	char keyFileName[800];
	struct stat keyFileNameInfo;
	int fd;
	size_t pos;
	u_int32_t semiflag;

	pos=registryGetFilename(key,keyFileName,sizeof(keyFileName));
	if (!pos) return -1; /* something is wrong */

	stat(keyFileName,&keyFileNameInfo);
	keyFromStat(key,&keyFileNameInfo);
	
	/* Remove the NEEDSYNC flag */
	semiflag=RG_KEY_FLAG_NEEDSYNC;
	semiflag=~semiflag;
	key->flags &= semiflag;

	return 0;
}



int registryGetKey(Key *key) {
	char keyFileName[500];
	struct stat keyFileNameInfo;
	int fd;
	size_t pos;
	u_int32_t semiflag;

	pos=registryGetFilename(key,keyFileName,sizeof(keyFileName));
	if (!pos) return -1; /* something is wrong */

	if ((fd=open(keyFileName,O_RDONLY))==-1) return -1;
	fstat(fd,&keyFileNameInfo);
	keyFromStat(key,&keyFileNameInfo);
	if (!keyIsDir(key)) {
		FILE *input;

		input=fdopen(fd,"r");
		if (keyFileUnserialize(key,input)) {
			fclose(input);
			return -1;
		}
		fclose(input);
	} else close(fd);

	/* Remove the NEEDSYNC flag */
	semiflag=RG_KEY_FLAG_NEEDSYNC;
	semiflag=~semiflag;
	key->flags &= semiflag;
	
	return 0;
}



int registrySetKeys(KeySet *ks) {
	Key *current;
	
	for (current=ks->start; current; current=current->next) {
		if (keyNeedsSync(current)) {
			registrySetKey(current);
		}
	}
	
	return 0;
}



int registrySetKey(Key *key) {
	char keyFileName[800];
	char folderMaker[800];
	char *cursor, *last;
	int fd;
	FILE *output=0;
	void *buffer;
	size_t pos,size;
	u_int32_t semiflag;
	struct stat stated;
	struct passwd *user=0;

	pos=registryGetFilename(key,keyFileName,sizeof(keyFileName));
	if (!pos) return -1; /* something is wrong */

	if (stat(keyFileName,&stated))
		if (errno==ENOENT) {
			/* check if parent dir already exists */
			last=rindex(keyFileName,'/');
			strncpy(folderMaker,keyFileName,last-keyFileName);
			folderMaker[last-keyFileName]=0;
			if (stat(folderMaker,&stated)) {
				/* create all path recursively until before our basename */
				last   =rindex(keyFileName,'/');
				cursor = index(keyFileName,'/'); cursor++; /* skip first occurence */
				if (!last || !cursor) {
					errno=RG_KEY_RET_INVALIDKEY;
					return -1;
				}
				for (cursor=index(cursor,'/');
						cursor && (cursor <= last);
						cursor=index(cursor,'/')) {
					strncpy(folderMaker,keyFileName,cursor-keyFileName);
					folderMaker[cursor-keyFileName]=0;
					/* TODO: use correct user's umask() for dir permissions */
					if (mkdir(folderMaker,0775)<0 && errno!=EEXIST) return -1;
					cursor++;
				}
			}
		} else return -1;
	else {
		/* A file or dir is already there. Lets check details */
		if ( S_ISDIR(stated.st_mode) && !keyIsDir(key)) {
			errno=EISDIR;
			return -1;
		}
		if (!S_ISDIR(stated.st_mode) &&  keyIsDir(key)) {
			errno=ENOTDIR;
			return -1;
		}
	}

	if (!keyIsDir(key)) {
		/* Try to open key file with its full file name */
		fd=open(keyFileName,O_CREAT | O_RDWR | O_TRUNC, keyGetAccess(key));
		if (fd==-1) return -1;
		if (getuid() == 0) fchown(fd,keyGetUID(key),keyGetGID(key));
		if (!(output=fdopen(fd,"w+"))) return -1;
		if (keyFileSerialize(key,output)) {
			fclose(output);
			return -1;
		}
		fclose(output);
	} else if (mkdir(keyFileName,keyGetAccess(key))<0 && errno!=EEXIST) return -1;

	/* Remove the NEEDSYNC flag */
	semiflag=RG_KEY_FLAG_NEEDSYNC;
	semiflag=~semiflag;
	key->flags &= semiflag;
	
	return 0;
}






int registryRemoveKey(char *keyName) {
	Key key;
	char fileName[800];
	off_t rc;

	keyInit(&key);
	rc=keySetName(&key,keyName);
	if (rc==-1) return -1;
	rc=registryGetFilename(&key,fileName,sizeof(fileName));
	if (!rc) return -1;

	return remove(fileName);
}


