/***************************************************************************
                          registry.h  -  Methods for accessing the Linux Registry
                             -------------------
    begin                : Mon Dec 29 2003
    copyright            : (C) 2003 by Avi Alkalay
    email                : avi@unix.sh
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef REGISTRY_H
#define REGISTRY_H

#include <sys/types.h>
#include <errno.h>

/* Registry data directories */
#define RG_DB_SYSTEM      "/etc/registry"
#define RG_DB_USER        ".registry"   /* $HOME/.registry */
#define RG_KEY_FORMAT     "RG001"



/* Return codes generated by class Registry methods */
#define RG_REG_RET_

/****************************************************/
/* Key data types */
#define RG_KEY_TYPE_UNDEFINED 0
#define RG_KEY_TYPE_BINARY    1
#define RG_KEY_TYPE_STRING    2
#define RG_KEY_TYPE_DIR       0xff

/* future use */
#define RG_KEY_TYPE_LINK      0xfe

/* deprecated */
#define RG_KEY_TYPE_DWORD     3
#define RG_KEY_TYPE_DOUBLE    4





/****************************************************/
/* Key name space types */
#define RG_NS_SYSTEM        1  /* The 'system.*' keys */
#define RG_NS_USER          2  /* The   'user.*' keys */




/* Numerical types (deprecated) */
#define RG_DWORD  long long int
#define RG_DOUBLE double


/*
The key flags bit array. The '.' indicator means unknown:

7654 3210 7654 3210 7654 3210 7654 3210
...1 0... .0.. .1.. ..1. ..0. ...0 1... 0x10042008 Initialized
...1 1... .1.. .1.. ..1. ..1. ...1 1... 0x18442218 Initialized mask
.... .... .... .... .... .... .... ...1 0x00000001 HASTYPE
.... .... .... .... .... .... .... ..1. 0x00000002 HASKEY
.... .... .... .... .... .... .... .1.. 0x00000004 HASDATA
.... .... .... .... .... .... ..1. .... 0x00000020 HASDOMAIN
.... .... .... .... .... .... .1.. .... 0x00000040 HASCOMMENT
.... .... .... .... .... .... 1... .... 0x00000060 HASUID
.... .... .... .... .... ...1 .... .... 0x00000080 HASGID
.... .... .... .... .... .1.. .... .... 0x00000400 HASPRM
.... .... .... .... .... 1... .... .... 0x00000800 HASTIME
.... .... .... .... ...1 .... .... .... 0x00001000 NEEDSYNC
*/

/* Key flags */
#define RG_KEY_FLAG_INITIALIZED       0x10042008
#define RG_KEY_FLAG_INITMASK          0x18442218

#define RG_KEY_FLAG_HASTYPE           1
#define RG_KEY_FLAG_HASKEY            1<<1
#define RG_KEY_FLAG_HASDATA           1<<2
#define RG_KEY_FLAG_HASDOMAIN         1<<5
#define RG_KEY_FLAG_HASCOMMENT        1<<6
#define RG_KEY_FLAG_HASUID            1<<7
#define RG_KEY_FLAG_HASGID            1<<8
#define RG_KEY_FLAG_HASPRM            1<<10
#define RG_KEY_FLAG_HASTIME           1<<11
#define RG_KEY_FLAG_NEEDSYNC          1<<12


/* Return codes generated by Key class' methods */
#define RG_KEY_RET_OK                       0
#define RG_KEY_RET_NULLKEY                  EINVAL
#define RG_KEY_RET_UNINITIALIZED            EINVAL
#define RG_KEY_RET_NOKEY                    ENODATA
#define RG_KEY_RET_NODATA                   ENODATA
#define RG_KEY_RET_NOGROUP                  ENODATA
#define RG_KEY_RET_NODESC                   ENODATA
#define RG_KEY_RET_NODOMAIN                 ENODATA
#define RG_KEY_RET_NOCRED                   EACCES
#define RG_KEY_RET_NOTIME                   ENODATA
#define RG_KEY_RET_TRUNC                    ENOBUFS
#define RG_KEY_RET_TYPEMISMATCH             EILSEQ
#define RG_KEY_RET_INVALIDKEY               EAFNOSUPPORT
#define RG_KEY_RET_NOTFOUND                 ENOENT



/* Options to be ORed in some methods */
#define RG_O_RECURSIVE          1
#define RG_O_DIR                1<<1
#define RG_O_NOVALUE            1<<2
#define RG_O_NOEMPTY            1<<3
#define RG_O_STATONLY           1<<4



typedef struct _Key {
	 u_int8_t      type;        /* data type */
	    uid_t      uid;         /* owner user ID */
	    uid_t      gid;         /* owner group ID */
	   mode_t      access;      /* access control */
	   time_t      atime;       /* time for last access */
	   time_t      mtime;       /* time for last modidification */
	   time_t      ctime;       /* time for last change (meta info) */
	   size_t      commentSize; /* size of the description string */
	   size_t      dataSize;    /* size of the value */
	   size_t      recordSize;  /* dataSize + commentSize + control */
	u_int32_t      flags;       /* Some control flags */
	   char *      key;         /* The name of the key */
	   char *      comment;     /* A comment about this key-value pair */
	   char *      userDomain;  /* User domain */
	   void *      data;        /* The value */
	struct _Key *  next;
} Key;



/* Key Name Anatomy

Key::key is the key name. It is a unique identifier for a registry key.
An exmaple of a complex key name is:

	user:some\.user.My Environment.SOME\.VAR

From this example:
  Root name = "user:some\.user"
      Owner = "some.user"
  Base name = "SOME\.VAR"
Parent name = "user:some\.user.My Environment"

*/





typedef struct _KeySet {
	Key *          start;
	Key *          end;
	unsigned int   size;
} KeySet;





/**************************************

Registry methods

***************************************/

/** All registry operations must be done on an opened registry object.
This method initializes a connection to the Registry server */
int registryOpen();


/** Closes the connection to the Registry server, and frees some internal memory */
int registryClose();


int registryGetValue(char *keyname, char *returned,size_t maxSize);
int registrySetValue(char *keyname, char *value);

// int registryGetInteger(char *keyname, RG_DWORD *returned);
// int registrySetInteger(char *keyname, RG_DWORD value);
//
// int registryGetDouble(char *keyname, RG_DOUBLE *returned);
// int registrySetDouble(char *keyname, RG_DOUBLE value);
//
// int registryGetBinary(char *keyname, void *returned,int maxSize);
// int registrySetBinary(char *keyname, void *value,int size);

// int registryGetLink(char *keyname, char *returned,int maxSize);
// int registrySetLink(char *keyname, char *value);

int registryGetComment(char *keyname, char *returned,size_t maxSize);
size_t registrySetComment(char *keyname, char *comment);

int registryStatKey(Key *key);
int registryGetKey(Key *key);
int registrySetKey(Key *key);
int registryRemoveKey(char *keyName);

//int registryGetGroupOfKeys(char *root, char *groupName, KeySet *returned);
int registryGetChildKeys(char *parent, KeySet *returned,unsigned long options);
int registryGetRootKeys(KeySet *returned);
/* int registryStatChildKeys(char *parentName, KeySet *returned, int recursive); */

int registrySetKeys(KeySet *ks);





/**************************************

Key methods

***************************************/

int keyInit(Key *key);
int keyClose(Key *key);

int keyIsInitialized(Key *key);
int keyNeedsSync(Key *key);
int keyDup(Key *source,Key *dest);

int keySerialize(Key *key,void *buffer, size_t maxSize);
int keyUnserialize(Key *key,void *buffer);
size_t keyGetSerializedSize(Key *key);


u_int8_t keyGetType(Key *key);
u_int8_t keySetType(Key *key,u_int8_t type);


size_t keyGetRecordSize(Key *key);
size_t keyGetNameSize(Key *key);
size_t keyGetFullNameSize(Key *key);

size_t keyGetName(Key *key, char *returnedName, size_t maxSize);
size_t keySetName(Key *key, char *newName);

size_t keyGetFullName(Key *key, char *returnedName, size_t maxSize);
size_t keyGetRootName(Key *key, char *returned, size_t maxSize);
size_t keyGetFullRootName(Key *key, char *returned, size_t maxSize);

size_t keyGetBaseName(Key *key, char *returned, size_t maxSize);
size_t keyNameGetBaseNameSize(char *keyName);
size_t keyGetBaseNameSize(Key *key);

size_t keyNameGetRootNameSize(char *keyName);
size_t keyGetRootNameSize(Key *key);
size_t keyGetFullRootNameSize(Key *key);


size_t keyGetCommentSize(Key *key);
size_t keyGetComment(Key *key, char *returnedDesc, size_t maxSize);
size_t keySetComment(Key *key, char *newDesc);

uid_t keyGetUID(Key *key);
int keySetUID(Key *key, uid_t uid);

gid_t keyGetGID(Key *key);
int keySetGID(Key *key, gid_t gid);

mode_t keyGetAccess(Key *key);
int keySetAccess(Key *key, mode_t mode);

size_t keyGetOwner(Key *key, char *returned, size_t maxSize);
size_t keySetOwner(Key *key, char *userDomain);


size_t keyGetDataSize(Key *key);

// int keyGetInteger(Key *key, RG_DWORD *returnedInt);
// size_t keySetInteger(Key *key, RG_DWORD newInt);
//
// int keyGetDouble(Key *key, RG_DOUBLE *returnedDouble);
// size_t keySetDouble(Key *key, RG_DOUBLE newDouble);

size_t keyGetString(Key *key, char *returnedString, size_t maxSize);
size_t keySetString(Key *key, char *newString);

size_t keyGetBinary(Key *key, void *returnedBinary, size_t maxSize);
size_t keySetBinary(Key *key, void *newBinary, size_t dataSize);
size_t keySetRaw(Key *key, void *newBinary, size_t dataSize);


time_t keyGetMTime(Key *key);
time_t keyGetATime(Key *key);
time_t keyGetCTime(Key *key);

size_t keyGetParentName(Key *key, char *returned, size_t maxSize);

size_t keyToString(Key *key, char *returned, size_t maxSize);

int keyIsSystem(Key *key);
int keyNameIsSystem(char *keyName);

int keyIsUser(Key *key);
int keyNameIsUser(char *keyName);

int keyGetNameSpace(Key *key);
int keyNameGetNameSpace(char *keyName);

int keyIsDir(Key *key);



/**************************************

KeySet methods

***************************************/

int ksInit(KeySet *ks);
int ksClose(KeySet *ks);

unsigned int ksInsert(KeySet *ks, Key *toInsert);
unsigned int ksAppend(KeySet *ks, Key *toAppend);

unsigned int ksInsertKeys(KeySet *ks, KeySet *toInsert);
unsigned int ksAppendKeys(KeySet *ks, KeySet *toAppend);

// Key *ksLookupByName(KeySet *ks,char *keyName);
// Key *ksLookupByRegex(KeySet *ks,regex_t *regex);




/***************************************

Helpers

***************************************/


size_t strblen(char *s);


/*

Example client code:

Registry registry;
int rgerror;
char buffer[100];
int someInteger;
double someDouble;
Key myKey;

registryOpen(&registry);
registryGetValue(registry,"system.userdb.root.group",buffer,sizeof(buffer));
registrySetValue(registry,"system.init.runlevel","5");

registryGetInteger(registry,"system.init.runlevel",&someInteger);
registrySetInteger(registry,"system.init.runlevel",3);

registryGetDouble(registry,"system.My Application.floatValue",&someDouble);
registrySetDouble(registry,"system.My Application.floatValue",3.14159);

registryGetBinary(registry,
	"user.identities.id0.Crypto Private Key",buffer,sizeof(buffer));
registrySetBinary(registry,
	"user.identities.id0.Crypto Private Key",buffer,binaryValueSize);

registryGetLink(registry,"system.userdb.500",buffer,sizeof(buffer));
buffer="system.My MP3 Encoder.quality";
registrySetLink(registry,"user.My MP3 Encoder.quality",buffer,strlen(buffer));


keyInit(&myKey);
keySetName(&myKey,"system.network.eth0.ip");

registryGetKey(registry,&myKey);

keySetString(&myKey,"192.168.10.20");

registrySetKey(registry,"user.myApp.Some Key",&myKey);

keyClose(&myKey); // to free data

registryClose(&registry); // to free data

*/





#endif /* REGISTRY_H */
