/***************************************************************************
                          rg.c  -  Tool for Registry administration
                             -------------------
    begin                : Mon Mar 02 2003
    copyright            : (C) 2003 by Avi Alkalay
    email                : avi@unix.sh
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <grp.h>
#include <pwd.h>
#include <time.h>

#include "registry.h"

/*****************************

# Get the value of a key
rg get system.some.key



# Set the value. Tries to keep current type. If it can't change it.
rg set user.some.key=value
rg set user.some.key -- value



# Set type, comment, user, group and value
rg set -t type -c "Some comment" -u user -g group user.some.key -- value



# Set key from standard input, trying to keep type
echo "value" | rg set user.some.key -



# Set the key in the domain of 'someuser', overwriting the user domain embedded in key name.
# Usefull in scripts.
rg set -d someuser user:otheruser.some.key=value



# Remove key/keys
rg rm system.some.key
rg rm -rf system.some.directory




# List key. If it is a directory, list the child keys.
# -l is for long listing, -v is to show the values.
rg ls user.some.directory
rg ls -l user.some.directory
rg ls user.some.key
rg ls -l user.some.key
rg ls -lv user:luciana.some.directory

*/


#define CMD_GET       1
#define CMD_SET       2
#define CMD_REMOVE    3
#define CMD_LIST      4

#define ARGSIZE      30

char *argComment=0;
char *argData=0;
char *argKeyName=0;
char *argDomain=0;
uid_t argUID=0;
uid_t argGID=0;
int argCommand=0;
int argRecursive=0;
int argLong=0;
int argValue=0;
int argDescriptive=0;
int argFullName=0;
mode_t argMode=0;
int argType=RG_KEY_TYPE_UNDEFINED;


int parseCommandLine(int argc, char *argv[]) {
	char sargType[ARGSIZE],argUser[ARGSIZE],argGroup[ARGSIZE];
	char sargMode[ARGSIZE],sargCommand[ARGSIZE];

	int opt;

	int test;

	*sargType=*argUser=*argGroup=*sargCommand=*sargMode=0;

	while ((opt=getopt(argc,argv,"-t:c:u:g:m:rflvRd"))!=-1) {
		switch (opt) {
			case 't':
				strncpy(sargType,optarg,ARGSIZE);
				break;
			case 'c':
				argComment=realloc(argComment,strlen(optarg)+1);
				strcpy(argComment,optarg);
				break;
			case 'u':
				strncpy(argUser,optarg,ARGSIZE);
				break;
			case 'g':
				strncpy(argGroup,optarg,ARGSIZE);
				break;
			case 'm':
				strncpy(sargMode,optarg,ARGSIZE);
				break;
			case 'R':
				argRecursive=RG_O_RECURSIVE;
				break;
			case 'l':
				argLong=1;
				break;
			case 'v':
				argValue=1;
				break;
			case 'd':
				argDescriptive=1;
				argLong=1;
				break;
			case 'f':
				argFullName=1;
				break;
			case 1: { /* handle non '-x' args */
				test=optind;
				if (*sargCommand==0) { /* parse sub-command */
					strncpy(sargCommand,optarg,ARGSIZE);
				} else if (!argKeyName) { /* parse key name */
					argKeyName=realloc(argKeyName,strlen(optarg)+1);
					strcpy(argKeyName,optarg);
				} else if (!argData) { /* parse value */
					argData=realloc(argData,strlen(optarg)+1);
					strcpy(argData,optarg);
				}
				break;
			}
		}
	}
	test=optind;

	/* Now check if we have the '--' argument, and get all its values */
	if (!strcmp(argv[optind-1],"--") && optind<argc) {
		int wordind=optind;
		size_t valueLength=0;

		while (wordind<argc) valueLength+=strlen(argv[wordind++])+1;
		argData=realloc(argData,valueLength);
		strcpy(argData,argv[optind++]);
		/* very ugly */
		while (optind<argc) sprintf(argData,"%s %s",argData,argv[optind++]);
	}

	/* End of command line argument reading. Now parse and finalize */

	/* Check parsed command */
	if (!strcmp(sargCommand,"ls")) argCommand=CMD_LIST;
	else if (!strcmp(sargCommand,"set")) argCommand=CMD_SET;
	else if (!strcmp(sargCommand,"get")) argCommand=CMD_GET;
	else if (!strcmp(sargCommand,"rm")) argCommand=CMD_REMOVE;
	else {
		fprintf(stderr,"Invalid subcommand\n");
		exit(1);
	}

	/* Parse type */
	if (*sargType!=0) {
		/* TODO: use regex */
		if (!strcmp(sargType,"string")) argType=RG_KEY_TYPE_STRING;
		else if (!strcmp(sargType,"bin")) argType=RG_KEY_TYPE_BINARY;
		else if (!strcmp(sargType,"binary")) argType=RG_KEY_TYPE_BINARY;
		else if (!strcmp(sargType,"dir")) argType=RG_KEY_TYPE_DIR;
	} else if (argCommand==CMD_SET) { /* We must have a type */
		argType=RG_KEY_TYPE_STRING;
		/*  TODO: */
		/* check if it is a double */
		/* check if it is an integer */
	}

	/* TODO: Parse user */
	/* TODO: Parse group */

	/* Parse permissions */
	if (*sargMode!=0) argMode=strtol(sargMode,0,8);

	/* Semantic check */


	return argCommand;
}


void listAccess(mode_t mode,char *readable) {
	readable[0] = S_ISDIR(mode)  ? 'd' : '-';
	readable[1] = mode & S_IRUSR ? 'r' : '-';
	readable[2] = mode & S_IWUSR ? 'w' : '-';
	readable[3] = mode & S_IXUSR ? 'x' : '-';
	readable[4] = mode & S_IRGRP ? 'r' : '-';
	readable[5] = mode & S_IWGRP ? 'w' : '-';
	readable[6] = mode & S_IXGRP ? 'x' : '-';
	readable[7] = mode & S_IROTH ? 'r' : '-';
	readable[8] = mode & S_IWOTH ? 'w' : '-';
	readable[9] = mode & S_IXOTH ? 'x' : '-';
	readable[10]= 0;
}


void listTime(time_t when,char *readable) {
	time_t current_time=time(0);
	char buf[400];
	time_t six_months_ago;
	int recent;

	/* If the file appears to be in the future, update the current
	   time, in case the file happens to have been modified since
	   the last time we checked the clock.  */

	/* Consider a time to be recent if it is within the past six
	   months.  A Gregorian year has 365.2425 * 24 * 60 * 60 ==
	   31556952 seconds on the average.  Write this value as an
	   integer constant to avoid floating point hassles.  */
	six_months_ago = current_time - 31556952 / 2;
	recent = (six_months_ago <= when) && (when <= current_time);

	ctime_r(&when,buf); /* buf will become "Wed Jun 30 21:49:08 1993\n" */
	memcpy(readable,buf+4,7); /* take only month and day */
	if (recent) {
		memcpy(readable,buf+4,12);
		readable[12]=0;
	} else {
		memcpy(readable,buf+4,7);
		readable[7]=' ';
		memcpy(readable+8,buf+20,4);
		readable[12]=0;
	}
}



void listSingleKey(Key *key) {
	char buffer[400];
	char *p=buffer;

	if (argLong) {
		struct passwd *pwd;
		struct group *grp;

		listAccess(keyGetAccess(key),p);
		p+=strlen(p);
		*p=' '; p++;
		*p=' '; p++;
		*p=' '; p++;

		pwd=getpwuid(keyGetUID(key));
		strcpy(p,pwd->pw_name);
		p+=strlen(p);
		*p=' '; p++;
		*p=' '; p++;

		grp=getgrgid(keyGetGID(key));
		strcpy(p,grp->gr_name);
		p+=strlen(p);
		*p=' '; p++;

		sprintf(p,"%*d ",5,keyGetRecordSize(key));
		p+=strlen(p);

		listTime(keyGetMTime(key),p);
		p+=strlen(p);
		*p=' '; p++;
	}
	keyGetFullName(key,p,sizeof(buffer)-(p-buffer));
	if (argValue && (keyGetDataSize(key)>0)) {
		p+=strlen(p);
		*p='='; p++;
		p+=keyGetString(key,p,sizeof(buffer)-(p-buffer));
		*p=0;
	}
	puts(buffer);
}








int commandRemove() {
	if (!argKeyName) {
		fprintf(stderr,"rg rm: No key name\n");
		return -1;
	}

	if (registryRemoveKey(argKeyName)) {
		perror("rg rm");
		return -1;
	}
	return 0;
}







int commandSet() {
	Key key;
	int ret;
	char error[200];


	/* Consistency */
	if (!argKeyName) {
		fprintf(stderr,"rg set: No key name\n");
		return -1;
	}

	keyInit(&key);
	keySetName(&key,argKeyName);
	ret=registryGetKey(&key);
	if (!ret) { /* key already exist. good. */
		if (argComment) keySetComment(&key,argComment);
		if (argType==RG_KEY_TYPE_UNDEFINED) argType=keyGetType(&key);
	} else if (errno!=RG_KEY_RET_NOTFOUND) {
		sprintf(error,"rg set: %s",argKeyName);
		perror(error);
	}

	/* TODO: handle UID and GID */

	if (argMode) keySetAccess(&key,argMode);

	switch (argType) {
		case RG_KEY_TYPE_DIR: keySetType(&key,RG_KEY_TYPE_DIR);
			break;
		case RG_KEY_TYPE_STRING: keySetString(&key,argData);
			break;
		case RG_KEY_TYPE_BINARY: keySetBinary(&key,argData,strlen(argData));
			break;
	}

	ret=registrySetKey(&key);
	if (ret) {
		sprintf(error,"rg set: %s",argKeyName);
		perror(error);
	}
	return ret;
}














int commandList() {
	Key *key=0;
	KeySet ks;
	int ret;

	ksInit(&ks);
	
	if (!argKeyName) {
		KeySet roots;
		/* User don't want a specific key, so list the root keys */
		
		ksInit(&roots);
		registryGetRootKeys(&roots);
		
		if (argRecursive) {
			key=roots.start;
			while (key) {
				char rootName[200];
				KeySet thisRoot;
				Key *temp;
				
				ksInit(&thisRoot);
				keyGetFullName(key,rootName,sizeof(rootName));
				if (argValue) ret=registryGetChildKeys(rootName,&thisRoot,
					argRecursive | (argDescriptive?RG_O_DIR:0));
				else ret=registryGetChildKeys(rootName,&thisRoot,
					RG_O_STATONLY | argRecursive | (argDescriptive?RG_O_DIR:0));
				temp=key->next;
				ksAppend(&ks,key);
				ksAppendKeys(&ks,&thisRoot);
				key=temp;
			}
		} else ksAppendKeys(&ks,&roots);
	} else {
		/* User gave us a specific key to start with */

		if (argValue) ret=registryGetChildKeys(argKeyName,&ks,
			argRecursive | (argDescriptive?RG_O_DIR:0));
		else ret=registryGetChildKeys(argKeyName,&ks,
			RG_O_STATONLY | argRecursive | (argDescriptive?RG_O_DIR:0));
	
		if (ret) {
			/* We got an error. Check if it is because its not a folder key */
			if (errno==ENOTDIR) {
				/* We still have a chance, since there is something there */
				key=(Key *)malloc(sizeof(Key));
				keyInit(key);
				keySetName(key,argKeyName);
				if (argValue) ret=registryGetKey(key);
				else ret=registryStatKey(key);
				if (ret) {
					char error[200];
					
					sprintf(error,"rg ls: %s",argKeyName);
					perror(error);
					return ret;
				}
			} else { /* A real error */
				char error[200];
				
				sprintf(error,"rg ls: %s",argKeyName);
				perror(error);
				return ret;
			}
		}
	}

	if (key) listSingleKey(key);
	else if (ks.size)
		for(key=ks.start; key; key=key->next)
			listSingleKey(key);

	ksClose(&ks);
	keyClose(key);
	free(key);
	return 0;
}



int commandGet() {
	int ret;
	Key key;
	char *buffer;
	char *p;
	size_t size,cs=0;

	if (!argKeyName) {
		fprintf(stderr,"rg get: No key name\n");
		return -1;
	}

	keyInit(&key);
	keySetName(&key,argKeyName);

	ret=registryGetKey(&key);
	if (ret) {
		char error[200];
			
		sprintf(error,"rg get: %s",argKeyName);
		perror(error);
		return ret;
	}
	size=keyGetDataSize(&key);
	if (argDescriptive) {
		cs=keyGetCommentSize(&key);
		if (cs) size+=cs+3;
	}
	if (argLong) {
		if (argFullName) size+=keyGetFullNameSize(&key);
		else size+=keyGetNameSize(&key);
	}


	if (keyGetType(&key)<=RG_KEY_TYPE_BINARY) p=buffer=malloc(size+1);
	else p=buffer=malloc(size);


	if (argDescriptive) {
		if (cs) {
			p+=sprintf(p,"# ");
			p+=keyGetComment(&key,p,size-(p-buffer));
			*p='\n'; p++;
		}
	}
	if (argLong) {
		if (argFullName) p+=keyGetFullName(&key,p,size-(p-buffer));
		else p+=keyGetName(&key,p,size-(p-buffer));
		*p='='; p++;
	}

	keyGetString(&key,p,size-(p-buffer));
	if (keyGetType(&key)<=RG_KEY_TYPE_BINARY) {
		p+=keyGetDataSize(&key);
		*p=0;
	}

	printf("%s\n",buffer);
	free(buffer);
	
	return 0;
}



int doCommand(int command) {
	switch (command) {
		case CMD_SET: return commandSet();
		case CMD_LIST: return commandList();
		case CMD_GET: return commandGet();
		case CMD_REMOVE: return commandRemove();
	}
	return 0;
}


int main(int argc, char **argv) {
	int command=0;
	int ret=0;

	command=parseCommandLine(argc,argv);

	registryOpen();
	ret=doCommand(command);
	registryClose();

	return ret;
}

